"""
torchortho - CUDA-Accelerated Hermite Polynomial Computation

Author: Ismail Khalfaoui-Hassani
Date: February 2025
License: GPL-3.0
Description: Implements CUDA kernels for computing Hermite polynomials used in learnable activations.

For reproducing experiments:
- Vision: https://github.com/K-H-Ismail/ConvNeXt-ortho
- Language: https://github.com/K-H-Ismail/pytorch-language-models
"""

import torch
from torch import Tensor
from torch.autograd import Function

try:
    from numba import cuda

    @cuda.jit("void(float32[:],int64,float32[:])")
    def hermite_forward_cuda_kernel(x, n, out):
        """CUDA kernel to compute Hermite polynomials up to degree n."""
        idx = cuda.grid(1)
        if idx < x.size:
            base_idx = idx * n
            value = x[idx]
            out[base_idx] = 1.0  # He_0(x) = 1
            if n > 1:
                out[base_idx + 1] = value  # He_1(x) = x
            for k in range(2, n):
                out[base_idx + k] = (
                    value * out[base_idx + k - 1] - (k - 1) * out[base_idx + k - 2]
                )

    @cuda.jit("void(float32[:], int64, float32[:], float32[:])")
    def hermite_backward_cuda_kernel(x, n, out, grad_out):
        """CUDA kernel to compute Hermite polynomial gradients."""
        idx = cuda.grid(1)
        size = grad_out.size
        if idx < size:
            base_idx = idx * n
            grad = 0.0
            for k in range(1, n):
                grad += x[base_idx + k] * k * out[base_idx + k - 1]
            grad_out[idx] = grad

except Exception as _:
    pass


class HermitePolynomialsNumba(Function):
    @staticmethod
    def forward(ctx, input: Tensor, degree: int) -> Tensor:
        """Computes Hermite polynomials via CUDA or CPU."""
        ctx.degree = degree
        ctx.input_size = input.size()
        ctx.numel = input.numel()
        output = torch.zeros(*input.size(), degree + 1).to(input)

        threads_per_block = 2**16
        blocks_per_grid = (ctx.numel + threads_per_block - 1) // threads_per_block

        hermite_forward_cuda_kernel[threads_per_block, blocks_per_grid](
            input.detach().flatten(),
            degree + 1,
            output.detach().flatten(),
        )

        ctx.output = output
        return output.view(*input.size(), -1)

    @staticmethod
    def backward(ctx, grad: Tensor) -> tuple:
        """Computes gradient of Hermite polynomials."""
        threads_per_block = 2**16
        blocks_per_grid = (ctx.numel + threads_per_block - 1) // threads_per_block

        grad_output = torch.zeros(*ctx.input_size).to(grad)
        hermite_backward_cuda_kernel[threads_per_block, blocks_per_grid](
            grad.detach().flatten(),
            ctx.degree + 1,
            ctx.output.detach().flatten(),
            grad_output.detach().flatten(),
        )
        return (grad_output, None)


hermite_polynomials_numba = HermitePolynomialsNumba.apply



"""
torchortho - Learnable Polynomial, Trigonometric, and Tropical Activations

Author: Ismail Khalfaoui-Hassani
Date: February 2025
License: GPL-3.0
Description: This module initializes the torchortho package and imports activation functions.

For reproducing experiments:
- Vision: https://github.com/K-H-Ismail/ConvNeXt-ortho
- Language: https://github.com/K-H-Ismail/pytorch-language-models
"""

from .hermite_activation import HermiteActivation
from .fourier_activation import FourierActivation
from .tropical_activation import TropicalActivation, TropicalRationalActivation
from .cuda import hermite_polynomials_numba

__all__ = [
    "HermiteActivation",
    "FourierActivation",
    "hermite_polynomials_numba",
    "TropicalActivation",
    "TropicalRationalActivation",
]



"""
torchortho - Learnable Tropical and Tropical Rational Activation Functions

Author: Ismail Khalfaoui-Hassani
Date: February 2025
License: GPL-3.0
Description: Implements Tropical polynomial and rational activations.

For reproducing experiments:
- Vision: https://github.com/K-H-Ismail/ConvNeXt-ortho
- Language: https://github.com/K-H-Ismail/pytorch-language-models
"""

import math
import os
import pickle

import torch
import torch.nn as nn


class TropicalActivation(nn.Module):
    def __init__(self, degree):
        """
        Initializes the Tropical polynomial activation function.

        """
        super(TropicalActivation, self).__init__()
        self.degree = degree
        self.register_buffer("powers", torch.arange(0, degree + 1))
        self.coefficients = nn.Parameter(torch.ones(degree + 1))

    def forward(self, x):
        """
        Forward pass for the Tropical activation function.

        Parameters:
        x (torch.Tensor): Input tensor.

        Returns:
        torch.Tensor: Output tensor after applying Tropical activation function.
        """
        return (math.sqrt(2) / self.degree) * (
            x.unsqueeze(-1) * self.powers + self.coefficients
        ).max(-1)[0]


class TropicalRationalActivation(nn.Module):
    def __init__(self, degree, act_init=None, requires_grad=True, load_cached=True):
        """
        Initializes the Tropical rational activation function.

        """
        super(TropicalRationalActivation, self).__init__()
        self.degree = degree
        self.requires_grad = requires_grad
        self.cache_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            f"cache_tropical_rational_{degree}.pkl",
        )
        self.load_cached = load_cached
        self.powers_num = nn.Parameter(
            torch.randn(degree + 1), requires_grad=requires_grad
        )
        self.powers_denom = nn.Parameter(
            torch.randn(degree + 1), requires_grad=requires_grad
        )
        self.coefficients_num = nn.Parameter(
            torch.randn(degree + 1), requires_grad=requires_grad
        )
        self.coefficients_denom = nn.Parameter(
            torch.randn(degree + 1), requires_grad=requires_grad
        )
        if act_init:
            self.act_init = act_init
            self.init_coeffs()

    def init_coeffs(self):
        # Check for cached results
        if os.path.exists(self.cache_file) and self.load_cached:
            with open(self.cache_file, "rb") as f:
                cached_data = pickle.load(f)
                print("Loaded cached coefficients")
                self.coefficients_num = nn.Parameter(
                    cached_data["coefficients_num"], requires_grad=self.requires_grad
                )
                self.coefficients_denom = nn.Parameter(
                    cached_data["coefficients_denom"], requires_grad=self.requires_grad
                )
                self.powers_num = nn.Parameter(
                    cached_data["powers_num"], requires_grad=self.requires_grad
                )
                self.powers_denom = nn.Parameter(
                    cached_data["powers_denom"], requires_grad=self.requires_grad
                )
                return
        # Compute the coefficients and frequencies at init with a Hermite interpolation
        lim = self.degree
        act = self.act_init
        coefficients_num = nn.Parameter(
            self.coefficients_num.clone() + torch.randn_like(self.coefficients_num)
        )
        coefficients_denom = nn.Parameter(
            self.coefficients_denom.clone() + torch.randn_like(self.coefficients_denom)
        )
        powers_num = nn.Parameter(
            self.powers_num.clone() + torch.randn_like(self.powers_num)
        )
        powers_denom = nn.Parameter(
            self.powers_denom.clone() + torch.randn_like(self.powers_denom)
        )

        def model():
            x = torch.linspace(
                torch.tensor(-lim), torch.tensor(lim), 2**8
            )  # (torch.rand(2**8) - 0.5) * 2 * lim
            x.requires_grad = True
            y = act(x)
            # Compute the derivative of the activation with respect to x
            y_deriv = torch.autograd.grad(
                outputs=y,
                inputs=x,
                grad_outputs=torch.ones_like(x),
                create_graph=True,
            )[0]
            x.requires_grad = False

            x_num = x.unsqueeze(-1) * powers_num
            x_denom = x.unsqueeze(-1) * powers_denom
            x = (x_num + coefficients_num).max(-1)[0] - (
                x_denom + coefficients_denom
            ).max(-1)[0]

            x_deriv = None  # self.powers[index].to(x)

            return x, x_deriv, y, y_deriv

        loss_fn = nn.MSELoss()  # Mean Squared Error Loss
        optimizer = torch.optim.Adam(
            [coefficients_num, coefficients_denom, powers_num, powers_denom], lr=0.01
        )
        loss_min = 1e8
        (
            coefficients_num_min,
            coefficients_denom_min,
            powers_num_min,
            powers_denom_min,
        ) = (
            None,
            None,
            None,
            None,
        )

        # Training loop
        num_epochs = 40000
        print("Init activation coefficients")
        for _ in range(num_epochs):
            # Forward pass
            y_pred, y_pred_deriv, y, y_deriv = model()

            # Compute loss
            loss = loss_fn(y_pred, y)  # + loss_fn(y_pred_deriv, y_deriv)

            if loss < loss_min:
                loss_min = loss
                coefficients_num_min = coefficients_num.clone().detach()
                coefficients_denom_min = coefficients_denom.clone().detach()
                powers_num_min = powers_num.clone().detach()
                powers_denom_min = powers_denom.clone().detach()

            if loss < 1e-4:
                coefficients_num_min = coefficients_num.clone().detach()
                coefficients_denom_min = coefficients_denom.clone().detach()
                powers_num_min = powers_num.clone().detach()
                powers_denom_min = powers_denom.clone().detach()
                break
            # Backpropagation and optimization
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        print("loss_min: ", loss_min)
        # Cache the results
        with open(self.cache_file, "wb") as f:
            pickle.dump(
                {
                    "coefficients_num": coefficients_num_min,
                    "coefficients_denom": coefficients_denom_min,
                    "powers_num": powers_num_min,
                    "powers_denom": powers_denom_min,
                },
                f,
            )
            print("Cached coefficients to", self.cache_file)
        (
            self.coefficients_num,
            self.coefficients_denom,
            self.powers_num,
            self.powers_denom,
        ) = (
            nn.Parameter(coefficients_num_min, requires_grad=self.requires_grad),
            nn.Parameter(coefficients_denom_min, requires_grad=self.requires_grad),
            nn.Parameter(powers_num_min, requires_grad=self.requires_grad),
            nn.Parameter(powers_denom_min, requires_grad=self.requires_grad),
        )

    def forward(self, x):
        """
        Forward pass for the Tropical activation function.

        Parameters:
        x (torch.Tensor): Input tensor.

        Returns:
        torch.Tensor: Output tensor after applying Tropical activation function.
        """
        x_num = x.unsqueeze(-1) * self.powers_num
        x_denom = x.unsqueeze(-1) * self.powers_denom
        x = (x_num + self.coefficients_num).max(-1)[0] - (
            x_denom + self.coefficients_denom
        ).max(-1)[0]
        return x
